@@ -96,7 +96,7 @@ scan_socket_dir (const char *dir, const char *prefix)
 		}
 		
 		/* Sort into some repeatable order */
-		if (!cur_dir || strcmp (cur_dir, name)) {
+		if (!cur_dir || strcmp (cur_dir, name) > 0) {
 			g_free (cur_dir);
 			cur_dir = name;
 		} else
@@ -300,6 +300,11 @@ giop_thread_free (GIOPThread *tdata)
 	giop_thread_list = g_list_remove (giop_thread_list, tdata);
 	G_UNLOCK (giop_thread_list);
 
+	g_mutex_free (tdata->lock);
+	g_cond_free (tdata->incoming);
+	g_queue_free (tdata->async_ents);
+	g_queue_free (tdata->request_queue);
+ 
 	g_free (tdata);
 }
 
@@ -561,8 +566,7 @@ giop_main_run (void)
 void
 giop_shutdown (void)
 {
-	if (!giop_thread_io ())
-		giop_connections_shutdown ();
+	link_connections_close ();
 
 	if (link_loop) /* break into the linc loop */
 		g_main_loop_quit (link_loop);
