@@ -15,8 +15,22 @@
 #include "poa-private.h"
 #include "orbit-poa.h"
 
-#define POA_LOCK(poa)   LINK_MUTEX_LOCK(poa->base.lock)
-#define POA_UNLOCK(poa) LINK_MUTEX_UNLOCK(poa->base.lock)
+#ifdef DEBUG_LOCKS
+#  define LOCK_DEBUG(a) g_printerr("%p: %6s file %s: line %d (%s)\n", \
+				   poa, a, __FILE__, __LINE__, __PRETTY_FUNCTION__);
+#else
+#  define LOCK_DEBUG(a)
+#endif
+
+#define POA_LOCK(poa)   G_STMT_START { \
+				LOCK_DEBUG("lock") \
+				LINK_MUTEX_LOCK(poa->base.lock); \
+			} G_STMT_END
+
+#define POA_UNLOCK(poa) G_STMT_START { \
+				LINK_MUTEX_UNLOCK(poa->base.lock); \
+				LOCK_DEBUG("unlock") \
+			} G_STMT_END
 
 static GMutex     *ORBit_class_assignment_lock    = NULL;
 static GHashTable *ORBit_class_assignments        = NULL;
@@ -59,10 +73,10 @@ static void               ORBit_POAObject_handle_request    (ORBit_POAObject
 							     GIOPRecvBuffer    *recv_buffer,
 							     CORBA_Environment *ev);
 
-static void               ORBit_POA_deactivate_object       (PortableServer_POA  poa,
-							     ORBit_POAObject     pobj,
-							     CORBA_boolean       do_etherealize,
-							     CORBA_boolean       is_cleanup);
+static void               ORBit_POA_deactivate_object_T   (PortableServer_POA  poa,
+							   ORBit_POAObject     pobj,
+							   CORBA_boolean       do_etherealize,
+							   CORBA_boolean       is_cleanup);
      
 /* PortableServer_Current interface */
 static void
@@ -150,6 +164,8 @@ ORBit_classinfo_register (PortableServer_ClassInfo *ci)
 			     (gpointer) ci->class_name, ci);
 }
 
+#define VEPV_CACHE_SIZE(c) (c)[0]
+
 void
 ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
 			   PortableServer_ServantBase *servant,
@@ -173,10 +189,11 @@ ORBit_skel_class_register (PortableServer_ClassInfo   *ci,
 	ORBit_classinfo_register (ci);
 	if (!ci->vepvmap) {
 		CORBA_unsigned_long offset;
+		CORBA_unsigned_long epvmap_size;
 
-		/* FIXME: we could propagate the size nicely here */
-		ci->vepvmap = g_new0 (ORBit_VepvIdx, 
-				      *(ci->class_id) + 1);
+		epvmap_size = *(ci->class_id) + 1;
+		ci->vepvmap = g_new0 (ORBit_VepvIdx, epvmap_size);
+		VEPV_CACHE_SIZE(ci->vepvmap) = epvmap_size;
 
 		ci->vepvmap[*(ci->class_id)] = class_offset;
 		
@@ -353,9 +370,9 @@ ORBit_POA_remove_child (PortableServer_POA poa,
 }
 
 static gboolean
-ORBit_POA_destroy (PortableServer_POA  poa,
-		   CORBA_boolean       etherealize_objects,
-		   CORBA_Environment  *ev)
+ORBit_POA_destroy_T (PortableServer_POA  poa,
+		     CORBA_boolean       etherealize_objects,
+		     CORBA_Environment  *ev)
 {
 	GPtrArray *adaptors;
 	int        numobjs;
@@ -369,12 +386,13 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	if (poa->life_flags & (ORBit_LifeF_Deactivating|ORBit_LifeF_Destroying))
 		return FALSE;	/* recursion */
 
-	/* FIXME: this needs some hard-core thread safety work ... */
-
 	poa->life_flags |= ORBit_LifeF_Destroying;
 
 	adaptors = poa->orb->adaptors;
 
+	/* FIXME: this lock thrash is really ugly */
+	LINK_MUTEX_LOCK (ORBit_RootObject_lifecycle_lock);
+
 	/* Destroying the children is tricky, b/c they may die
 	 * while we are traversing. We traverse over the
 	 * ORB's global list (rather than poa->child_poas) 
@@ -382,10 +400,25 @@ ORBit_POA_destroy (PortableServer_POA  poa,
 	for (i = 0; i < adaptors->len; i++) {
 		PortableServer_POA cpoa = g_ptr_array_index (adaptors, i);
 
-		if (cpoa && cpoa->parent_poa == poa)
-			ORBit_POA_destroy (cpoa, etherealize_objects, ev);
+
+		LINK_MUTEX_UNLOCK (ORBit_RootObject_lifecycle_lock);
+
+		if (cpoa && cpoa != poa) {
+			ORBit_RootObject_duplicate (cpoa);
+			POA_LOCK (cpoa);
+
+			if (cpoa->parent_poa == poa) 
+				ORBit_POA_destroy_T (cpoa, etherealize_objects, ev);
+
+			POA_UNLOCK (cpoa);
+			ORBit_RootObject_release (cpoa);
+		}
+
+		LINK_MUTEX_LOCK (ORBit_RootObject_lifecycle_lock);
 	}
 
+	LINK_MUTEX_UNLOCK (ORBit_RootObject_lifecycle_lock);
+
 	poa->default_servant = NULL;
 
 	if (g_hash_table_size (poa->child_poas) > 0 || poa->use_cnt ||
@@ -429,7 +462,7 @@ traverse_cb (PortableServer_ObjectId *oid,
 	if (pobj->use_cnt > 0)
 		info->in_use = TRUE;
 
-	ORBit_POA_deactivate_object (info->poa, pobj, info->do_etherealize, TRUE);
+	ORBit_POA_deactivate_object_T (info->poa, pobj, info->do_etherealize, TRUE);
 }
 
 static gboolean
@@ -933,10 +966,10 @@ ORBit_POA_activate_object_T (PortableServer_POA          poa,
  * etherialization and memory release will occur later.
  */
 static void
-ORBit_POA_deactivate_object (PortableServer_POA poa,
-			     ORBit_POAObject    pobj,
-			     CORBA_boolean      do_etherealize,
-			     CORBA_boolean      is_cleanup)
+ORBit_POA_deactivate_object_T (PortableServer_POA poa,
+			       ORBit_POAObject    pobj,
+			       CORBA_boolean      do_etherealize,
+			       CORBA_boolean      is_cleanup)
 {
 	PortableServer_ServantBase *servant = pobj->servant;
 
@@ -993,8 +1026,11 @@ ORBit_POA_deactivate_object (PortableServer_POA poa,
 			/* In theory, the finalize fnc should always be non-NULL;
 			 * however, for backward compat. and general extended
 			 * applications we dont insist on it. */
-			if (epv && epv->finalize)
+			if (epv && epv->finalize) {
+				POA_UNLOCK (poa);
 				epv->finalize (servant, ev);
+				POA_LOCK (poa);
+			}
 		}
 		pobj->use_cnt--; /* allow re-activation */
 		g_assert (ev->_major == 0);
@@ -1103,8 +1139,10 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	ORBitSmallSkeleton                   small_skel = NULL;
 	gpointer                             imp = NULL;
 
-	if (poa)
+	if (poa) {
+		ORBit_RootObject_duplicate (poa);
 		POA_LOCK (poa);
+	}
 
 	if (!poa || !poa->poa_manager)
 		CORBA_exception_set_system (
@@ -1120,9 +1158,7 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 						
 				poa->held_requests = g_slist_prepend (
 					poa->held_requests, recv_buffer);
-				
-				POA_UNLOCK (poa);
-				return;
+				goto clean_out;
 			} else
 				CORBA_exception_set_system (
 					ev, ex_CORBA_TRANSIENT,
@@ -1180,12 +1216,6 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		}
 	}
 
-	pobj->use_cnt++;
-	LINK_MUTEX_LOCK (poa->orb->lock);
-	poa->orb->current_invocations =
-		g_slist_prepend (poa->orb->current_invocations, pobj);
-	LINK_MUTEX_UNLOCK (poa->orb->lock);
-	
 	if (ev->_major == CORBA_NO_EXCEPTION && !pobj->servant)
 		CORBA_exception_set_system (
 			ev, ex_CORBA_OBJECT_NOT_EXIST, 
@@ -1196,6 +1226,12 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		goto clean_out;
 	}
 
+	pobj->use_cnt++;
+	LINK_MUTEX_LOCK (poa->orb->lock);
+	poa->orb->current_invocations =
+		g_slist_prepend (poa->orb->current_invocations, pobj);
+	LINK_MUTEX_UNLOCK (poa->orb->lock);
+	
 	klass = ORBIT_SERVANT_TO_CLASSINFO (pobj->servant);
 
 	if (klass->impl_finder)
@@ -1224,30 +1260,26 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		}
 	}
 
-	if (ev->_major != CORBA_NO_EXCEPTION) {
+	if (ev->_major != CORBA_NO_EXCEPTION)
 		return_exception (recv_buffer, m_data, ev);
-		goto clean_out;
+
+	else {
+		POA_UNLOCK (poa);
+
+		if (recv_buffer) {
+			struct ORBit_POA_invoke_data invoke_data;
+			
+			invoke_data.small_skel = small_skel;
+			invoke_data.imp        = imp;
+			
+			ORBit_small_invoke_adaptor ((ORBit_OAObject) pobj, recv_buffer,
+						    m_data, &invoke_data, ev);
+		} else
+			small_skel (pobj->servant, ret, args, ctx, ev, imp);
+		
+		POA_LOCK (poa);
 	}
 
-	ORBit_RootObject_duplicate (poa);
-	POA_UNLOCK (poa);
-
-	if (recv_buffer) {
-		struct ORBit_POA_invoke_data invoke_data;
-
-		invoke_data.small_skel = small_skel;
-		invoke_data.imp        = imp;
-
-		ORBit_small_invoke_adaptor (
-			(ORBit_OAObject) pobj, recv_buffer,
-			m_data, &invoke_data, ev);
-	} else
-		small_skel (pobj->servant, ret, args, ctx, ev, imp);
-
-	POA_LOCK (poa);
-	ORBit_RootObject_release (poa);
-
- clean_out:
 	if (recv_buffer)
 		CORBA_exception_free (ev);
 
@@ -1259,7 +1291,7 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 				oid, pobj->servant, ev);
 			break;
 		case PortableServer_USE_DEFAULT_SERVANT:
-			ORBit_POA_deactivate_object (poa, pobj, FALSE, FALSE);
+			ORBit_POA_deactivate_object_T (poa, pobj, FALSE, FALSE);
 			break;
 		default:
 			g_assert_not_reached ();
@@ -1267,7 +1299,6 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 		}
 
 	LINK_MUTEX_LOCK (poa->orb->lock);
-/*	g_assert ((ORBit_POAObject)poa->orb->current_invocations->data == pobj); */
 	poa->orb->current_invocations =
 		g_slist_remove (poa->orb->current_invocations, pobj);
 	LINK_MUTEX_UNLOCK (poa->orb->lock);
@@ -1276,7 +1307,9 @@ ORBit_POAObject_handle_request (ORBit_POAObject    pobj,
 	if (pobj->life_flags & ORBit_LifeF_NeedPostInvoke)
 		ORBit_POAObject_post_invoke (pobj);             
 
+ clean_out:
 	POA_UNLOCK (poa);
+	ORBit_RootObject_release (poa);
 }
 
 static ORBit_POAObject
@@ -1616,8 +1649,7 @@ ORBit_POA_ServantManager_unuse_servant (PortableServer_POA                    po
 }
 
 /*
- * This function is invoked from the generated stub code, after
- * invoking the servant method.
+ * Was this ever exposed / can we axe it ?
  */
 void
 ORBit_POAObject_post_invoke (ORBit_POAObject pobj)
@@ -1630,7 +1662,7 @@ ORBit_POAObject_post_invoke (ORBit_POAObject pobj)
 		 * are stored in pobj->life_flags and they dont need
 		 * to be passed in again!
 		 */
-		ORBit_POA_deactivate_object (
+		ORBit_POA_deactivate_object_T (
 			pobj->poa, pobj, /*ether*/0, /*cleanup*/0);
 
 		/* WATCHOUT: pobj may not exist anymore! */
@@ -1813,18 +1845,23 @@ PortableServer_POA_destroy (PortableServer_POA   poa,
 
 	poa_sys_exception_if_fail (poa != NULL, ex_CORBA_INV_OBJREF);
 
-	if (poa->life_flags & ORBit_LifeF_Destroyed)
-		return;
+	ORBit_RootObject_duplicate (poa);
+	POA_LOCK (poa);
 
-	if (wait_for_completion && ORBit_POA_is_inuse (poa, CORBA_TRUE, ev)) {
+	if (poa->life_flags & ORBit_LifeF_Destroyed)
+		;
+
+	else if (wait_for_completion && ORBit_POA_is_inuse_T (poa, CORBA_TRUE, ev))
 		CORBA_exception_set_system (ev, ex_CORBA_BAD_INV_ORDER,
 					    CORBA_COMPLETED_NO);
-		return;
+	else {
+		done = ORBit_POA_destroy_T (poa, etherealize_objects, ev);
+
+		g_assert (done || !wait_for_completion);
 	}
 
-	done = ORBit_POA_destroy (poa, etherealize_objects, ev);
-
-	g_assert (done || !wait_for_completion);
+	POA_UNLOCK (poa);
+	ORBit_RootObject_release (poa);
 }
 
 CORBA_string
@@ -2043,7 +2080,7 @@ PortableServer_POA_deactivate_object (PortableServer_POA             poa,
 	pobj = ORBit_POA_object_id_lookup_T (poa, oid);
 
 	if (pobj && pobj->servant)
-		ORBit_POA_deactivate_object (poa, pobj, CORBA_TRUE, CORBA_FALSE);
+		ORBit_POA_deactivate_object_T (poa, pobj, CORBA_TRUE, CORBA_FALSE);
 
 	POA_UNLOCK (poa);
 
@@ -2424,6 +2461,7 @@ get_c_method (CORBA_Object                 obj,
 
 	if (!obj ||
 	    !(pobj = (ORBit_POAObject) obj->adaptor_obj) ||
+	    !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
 	    !(*servant = (PortableServer_ServantBase *) pobj->servant))
 		return NULL;
 
@@ -2440,8 +2478,8 @@ get_c_method (CORBA_Object                 obj,
 	 * FIXME: we could propagate the size of the vepvmap_cache
 	 * to here, and check it to avoid some really bogus stuff.
 	 */
-	if (!class_id || !(pobj->base.interface->adaptor_type & ORBIT_ADAPTOR_POA) ||
-	    !pobj->vepvmap_cache)
+	if (!class_id || !pobj->vepvmap_cache ||
+	    class_id >= VEPV_CACHE_SIZE (pobj->vepvmap_cache) )
 		return NULL;
 
 	epv_start = (guchar *) (*servant)->vepv [ pobj->vepvmap_cache [class_id] ];
