@@ -127,6 +127,38 @@ testSequenceHelpers (void)
 		
 		CORBA_free (oseq);
 	}
+
+	/* test concat operation: concat two sequences of different
+	 * size. The concatinated sequence must contain values of both
+	 * in correct order. */
+	{
+		CORBA_octet oct = 0;
+		CORBA_long i = 0;
+		CORBA_long j = 0;
+		CORBA_long a = 0;
+		CORBA_long b = 0;
+		CORBA_sequence_CORBA_octet *aseq = NULL;
+		CORBA_sequence_CORBA_octet *bseq = NULL;
+
+		aseq = ORBit_sequence_alloc (TC_CORBA_sequence_CORBA_octet, 0);
+		bseq = ORBit_sequence_alloc (TC_CORBA_sequence_CORBA_octet, 0);
+		
+		for (b = 0; b < 200; ++b) {
+			ORBit_sequence_concat (aseq, bseq);
+			
+			a = 0;
+			for (i = 0; i < b; ++i)
+				for (j = 0; j < i; ++j, ++a)
+					g_assert (ORBit_sequence_index (aseq, a) == j % 128);
+			
+			oct = b % 128;
+			ORBit_sequence_append (bseq, &oct);
+		}
+		
+		CORBA_free (aseq);
+		CORBA_free (bseq);
+	}
+	
 }
 
 static void
@@ -434,10 +466,32 @@ static void
 testIsA (test_TestFactory   factory, 
 	 CORBA_Environment *ev)
 {
+	test_DerivedServer ds;
+ 
+	d_print ("Testing is_a ...\n");
+
 	g_assert (CORBA_Object_is_a (factory, "IDL:CORBA/Object:1.0", ev));
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
 	g_assert (CORBA_Object_is_a (factory, "IDL:omg.org/CORBA/Object:1.0", ev));
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	ds = test_TestFactory_getDerivedServer (factory, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	g_assert (CORBA_Object_is_a (ds, "IDL:orbit/test/DerivedServer:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (CORBA_Object_is_a (ds, "IDL:orbit/test/C1:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (CORBA_Object_is_a (ds, "IDL:orbit/test/B1:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (CORBA_Object_is_a (ds, "IDL:orbit/test/B2:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (CORBA_Object_is_a (ds, "IDL:orbit/test/BaseServer:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+	g_assert (CORBA_Object_is_a (factory, "IDL:CORBA/Object:1.0", ev));
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	CORBA_Object_release (ds, ev);
 }
 
 static void
@@ -1483,6 +1537,23 @@ testMisc (test_TestFactory   factory,
 	}
 }
 
+static void
+testLifeCycle (test_TestFactory   factory, 
+	       CORBA_Environment *ev)
+{
+	test_LifeCycleServer objref;
+
+	d_print ("Testing LifeCycle bits...\n");
+
+	objref = test_TestFactory_createLifeCycleServer (factory, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	test_LifeCycleServer_deactivateOnReturn (objref, ev);
+	g_assert (ev->_major == CORBA_NO_EXCEPTION);
+
+	CORBA_Object_release (objref, ev);
+}
+
 static volatile int done = 0;
 
 static void
@@ -2024,6 +2095,7 @@ run_tests (test_TestFactory   factory,
 		if (!in_proc)
 			testPingPong (factory, thread_tests, ev);
 		testMisc (factory, ev);
+		testLifeCycle (factory, ev);
 	}
 	
 #if NUM_RUNS > 1
@@ -2134,9 +2206,9 @@ main (int argc, char *argv [])
 	}
 
 	if (thread_safe)
-		orb_name = "orbit-local-mt-orb";
-	else
 		orb_name = "orbit-local-orb";
+	else
+		orb_name = "orbit-local-non-threaded-orb";
 
 	global_orb = CORBA_ORB_init (&argc, argv, orb_name, ev);
 	g_assert (ev->_major == CORBA_NO_EXCEPTION);
